// $Id: buffer.h,v 1.5 2008/11/20 11:14:20 dvermeir Exp $
#include <iostream>
#include <stdexcept>
#include <dvthread/lock.h>
#include <dvthread/thread.h>
#include <dvthread/monitor.h>

// A buffer with controlled access.
class Buffer: private Dv::Thread::Monitor {
  public:	
    // A buffer is a Monitor with 2 conditions.
    Buffer() throw (std::runtime_error): Monitor("buffer",2), n_items_(0) {}
  
     // Names for conditions. 
    enum { 
      OK_TO_GET = 0, // Signal if buffer not empty.
      OK_TO_PUT = 1   // Signal if buffer not full.  
    };
  
    void put(int i) throw (std::runtime_error) { 
      Dv::Thread::Lock lock(*this);
      while (n_items_==MAX) {
        std::cerr << Dv::Thread::Thread::self_id() 
          << ": Buffer::put: wait" << std::endl;
        if (! wait(OK_TO_PUT,2000) ) // Wait at most 2 secs.
	  throw std::runtime_error("Buffer::put() timed out"); 
      }
      std::cerr << Dv::Thread::Thread::self_id() 
        << ": Buffer::put: " << i << std::endl;
      data_[n_items_++] = i; // Actually put the item.
      signal(OK_TO_GET); 
    }

    int get() throw (std::runtime_error) { 
      Dv::Thread::Lock lock(*this);
      while (n_items_==0) {
        std::cerr << Dv::Thread::Thread::self_id() << ": Buffer::get: wait" << std::endl;
        if (!wait(OK_TO_GET, 2000))
	  throw std::runtime_error("Buffer::get() timed out");
      }
      int tmp = data_[--n_items_];
      std::cerr << Dv::Thread::Thread::self_id() << ": Buffer::get: " << tmp << std::endl;
      signal(OK_TO_PUT);
      return tmp;
    }
  private:
    enum { MAX = 3 }; // Buffer capacity.
    int n_items_; // Number of items in the buffer.
    int data_[MAX]; // Actual store for items.
};
